import torch
from .math_utils import affinity_from_features
from .nystrom_utils import farthest_point_sampling


@torch.no_grad()
def find_gamma_by_degree(features, degree, feature_B=None, 
                        distance='rbf', init_gamma=0.5, r_tol=1e-2, max_iter=100):
    # features: (n_samples, n_features)
    # binary search for optimal gamma, such that the mean edge weight is close to target_d
    gamma = init_gamma
    current_degree = affinity_from_features(features, features_B=feature_B, distance=distance, affinity_focal_gamma=gamma).mean()
    i_iter = 0
    low, high = 0, float('inf')
    tol = r_tol * degree
    while abs(current_degree - degree) > tol and i_iter < max_iter:
        if current_degree > degree:
            high = gamma
            gamma = (low + gamma) / 2
        else:
            low = gamma
            gamma = gamma * 2 if high == float('inf') else (gamma + high) / 2
        current_degree = affinity_from_features(features, features_B=feature_B, distance=distance, affinity_focal_gamma=gamma).mean()
        i_iter += 1
    return gamma

@torch.no_grad()
def find_gamma_by_degree_after_fps(features, degree, distance='rbf', init_gamma=0.5, r_tol=1e-2, max_iter=100, num_sample=1000):
    # features: (n_samples, n_features)
    # binary search for optimal gamma, such that the mean edge weight is close to target_d
    sample_indices = farthest_point_sampling(features, num_sample)
    sampled_features = features[sample_indices]
    return find_gamma_by_degree(sampled_features, degree, distance=distance, init_gamma=init_gamma, r_tol=r_tol, max_iter=max_iter)




import torch
from .nystrom_utils import farthest_point_sampling


# Multiclass Spectral Clustering, SX Yu, J Shi, 2003
def _discretisation_eigenvector(eigen_vector):
    # Function that discretizes rotated eigenvectors
    n, k = eigen_vector.shape

    # Find the maximum index along each row
    _, J = torch.max(eigen_vector, dim=1)
    Y = torch.zeros(n, k, device=eigen_vector.device).scatter_(1, J.unsqueeze(1), 1)

    return Y

@torch.no_grad()
def kway_ncut(eigen_vectors: torch.Tensor, 
              max_iter=1000, num_sample=10240,
              return_continuous=False,
              return_rotation=False):
    """Multiclass Spectral Clustering, SX Yu, J Shi, 2003

    Args:
        eigen_vectors (torch.Tensor): continuous eigenvectors from NCUT, shape (n, k)
        max_iter (int, optional): Maximum number of iterations.

    Returns:
        torch.Tensor: Discretized eigenvectors, shape (n, k), each row is a one-hot vector.
    """

    # Normalize eigenvectors
    vm = torch.sqrt(torch.sum(eigen_vectors ** 2, dim=1))
    eigen_vectors = eigen_vectors / vm.unsqueeze(1)

    # subsample the eigenvectors, to speed up the computation
    n, k = eigen_vectors.shape
    num_sample = max(num_sample, k)
    sample_idx = farthest_point_sampling(eigen_vectors, num_sample)
    _eigen_vectors = eigen_vectors[sample_idx]


    # Initialize R matrix with the first column from Farthest Point Sampling
    _sample_idx = farthest_point_sampling(_eigen_vectors, k)
    R = _eigen_vectors[_sample_idx].T
    
    # Iterative optimization loop
    last_objective_value = 0
    exit_loop = False
    nb_iterations_discretisation = 0

    while not exit_loop:
        nb_iterations_discretisation += 1

        # Discretize the projected eigenvectors
        _eigenvectors_continuous = _eigen_vectors @ R
        _eigenvectors_discrete = _discretisation_eigenvector(_eigenvectors_continuous)

        # SVD decomposition
        _out = _eigenvectors_discrete.T @ _eigen_vectors
        U, S, Vh = torch.linalg.svd(_out, full_matrices=False)
        V = Vh.T

        # Compute the Ncut value
        ncut_value = 2 * (n - torch.sum(S))

        # Check for convergence
        if torch.abs(ncut_value - last_objective_value) < torch.finfo(
            torch.float32).eps or nb_iterations_discretisation > max_iter:
            exit_loop = True
        else:
            last_objective_value = ncut_value
            R = V @ U.T

    eigenvectors_continuous = eigen_vectors @ R
    eigenvectors_discrete = _discretisation_eigenvector(eigenvectors_continuous)

    if return_rotation:
        return eigenvectors_discrete, R

    if return_continuous:
        return eigenvectors_continuous

    return eigenvectors_discrete


def axis_align(eigen_vectors, max_iter=1000, num_sample=10240):
    """deprecated, use kway_ncut(return_rotation=True) instead"""
    return kway_ncut(eigen_vectors, max_iter=max_iter, num_sample=num_sample, return_rotation=True)



# %%

from ncut_pytorch.new_ncut_pytorch import NewNCUT


from typing import List, Tuple
from ncut_pytorch import affinity_from_features, ncut, nystrom_ncut
import torch
import logging
import numpy as np
import matplotlib.pyplot as plt


def plot_eigvecs_dynamic(eigvec1, eigvec2, eigvec3, sample_index, x_2d, n_cols=6, tol=1e-3):
    n_eig = eigvec1.shape[1]

    n_rows = n_eig // n_cols * 3
    fig, axs = plt.subplots(n_rows, n_cols, figsize=(n_cols * 3, n_rows * 3))    
    
    # vmaxvmin_fn = lambda x: max(abs(x.min()), abs(x.max()))
    
    def vmaxvmin_fn(x):
        quantile = torch.tensor([0.01, 0.99])
        quantiles = x.flatten().quantile(quantile)
        return max(abs(quantiles[0]), abs(quantiles[1]))
        
    # plot one row of eigvec1, one row of eigvec2, repeat
    for i in range(n_eig):
        
        ax = axs[i // n_cols * 3, i % n_cols]
        _v = vmaxvmin_fn(eigvec1[:, i])
        sc = ax.scatter(x_2d[:, 0], x_2d[:, 1], c=eigvec1[:, i], cmap="bwr", vmin=-_v, vmax=_v)
        fig.colorbar(sc, ax=ax)
        ax.set_title(f"exact NCUT eig{i}")
            
        ax = axs[i // n_cols * 3 + 1, i % n_cols]
        _v = vmaxvmin_fn(eigvec2[:, i])
        sc = ax.scatter(x_2d[:, 0], x_2d[:, 1], c=eigvec2[:, i], cmap="bwr", vmin=-_v, vmax=_v)
        fig.colorbar(sc, ax=ax)
        ax.set_title(f"(wentinn) new nystrom NCUT eig{i}")
        
        ax = axs[i // n_cols * 3 + 2, i % n_cols]
        _v = vmaxvmin_fn(eigvec3[:, i])
        sc = ax.scatter(x_2d[:, 0], x_2d[:, 1], c=eigvec3[:, i], cmap="bwr", vmin=-_v, vmax=_v)
        fig.colorbar(sc, ax=ax)
        ax.set_title(f"(huzheng) bruteforce nystrom NCUT eig{i}")
        
    axs[0, 0].scatter(x_2d[sample_index, 0], x_2d[sample_index, 1], c="k", s=10)
    
    return fig

def non_appriximation_ncut(features, num_eig, **config_kwargs):
    aff = affinity_from_features(features, **config_kwargs)
    eigvec, eigval = ncut(aff, num_eig)
    return eigvec, eigval

from ncut_pytorch.new_ncut_pytorch import NewNCUT
def new_nystrom_ncut(features, num_eig, precomputed_sampled_indices, distance="rbf"):
    eigvec, eigval = NewNCUT(num_eig=num_eig, distance=distance).fit_transform(
        features, precomputed_sampled_indices=precomputed_sampled_indices)
    return eigvec, eigval

from ncut_pytorch.ncut_pytorch import NCUT
def knn_nystrom_ncut(features, num_eig, precomputed_sampled_indices, distance="rbf"):
    eigvec, eigval = NCUT(num_eig=num_eig, knn=10, distance=distance).fit_transform(
        features, precomputed_sampled_indices=precomputed_sampled_indices)
    return eigvec, eigval

from myold_nystrom import nystrom_ncut
from bruteforce_nystrom import force_nystrom_ncut


if __name__ == "__main__":
    

    from sklearn.datasets import make_blobs
    x_2d, y = make_blobs(n_samples=100, centers=2, n_features=2, random_state=42)
    x_2d = torch.tensor(x_2d, dtype=torch.float32)
    # sort by y
    x_2d = x_2d[y.argsort()]
    print(x_2d.shape)
    

    x_2d = torch.randn(100, 2)
    precomputed_sampled_indices = torch.arange(0, 12)
    
    num_eig = 10

    distance = "rbf"
    eigvec1, eigval1 = non_appriximation_ncut(x_2d, num_eig, distance=distance)
    
    eigvec2, eigval2 = new_nystrom_ncut(x_2d, num_eig, precomputed_sampled_indices, distance=distance)
    
    eigvec3, eigval3 = force_nystrom_ncut(x_2d, precomputed_sampled_indices, distance=distance, num_eig=num_eig)
    
    fig = plot_eigvecs_dynamic(eigvec1, eigvec2, eigvec3, 
                        precomputed_sampled_indices, x_2d, n_cols=num_eig,)
    fig.suptitle("""Comparison of exact NCUT, KNN nystrom NCUT, and original nystrom NCUT.
                 black dots are precomputed sampled indices""")
    fig.tight_layout()
    plt.show()


# %%



import torch

from .nystrom_utils import farthest_point_sampling, propagate_knn, which_device
from .affinity_gamma import find_gamma_by_degree_after_fps
from .math_utils import affinity_from_features, normalize_affinity, svd_lowrank, correct_rotation
from .kway_ncut import kway_ncut

def bias_ncut_soft(features, fg_idx, bg_idx=None,
                 num_eig=50, 
                 bias_factor=0.5, 
                 bg_factor=0.1,
                 num_sample=10240,
                 distance='rbf', degree=0.1, 
                 device=None):
    """
    Args:
        features: (n_nodes, n_features)
        fg_idx: (n_clicks) indices of the clicked points for the foreground
        bg_idx: (n_clicks) indices of the clicked points for the background
        bias_factor: (float) the factor of the bias term, decrease it to grow the mask bigger, need to be tuned for different images
        device: (torch.device)
        distance: (str) 'rbf' or 'cosine'
        degree: (float) the degree of the affinity matrix, 0.1 is good for most cases, decrease it will sharpen the affinity matrix
        num_sample: (int) increasing it does not necessarily improve the result
        num_eig: (int) does not matter since we only need a binary mask
    """
    if bg_idx is None:
        bg_idx = torch.tensor([], dtype=torch.long)

    n_nodes, n_features = features.shape
    num_sample = min(num_sample, n_nodes//4)
    # farthest point sampling
    fps_idx = farthest_point_sampling(features, num_sample=num_sample)
    fps_idx = torch.tensor(fps_idx, dtype=torch.long)
    # remove pos_idx and neg_idx from fps_idx
    fps_idx = fps_idx[~torch.isin(fps_idx, torch.cat([fg_idx, bg_idx]))]
    # add pos_idx and neg_idx to fps_idx
    fps_idx = torch.cat([fg_idx, bg_idx, fps_idx])
    fg_idx = torch.arange(len(fg_idx))
    bg_idx = torch.arange(len(bg_idx)) + len(fg_idx)
    
    device = which_device(features.device, device)
    _input = features[fps_idx].to(device)

    gamma = find_gamma_by_degree_after_fps(_input, degree=degree, distance=distance)
    affinity = affinity_from_features(_input, distance=distance, affinity_focal_gamma=gamma)
    affinity = normalize_affinity(affinity)
    
    # modify the affinity from the clicks
    click_f = 1 * affinity[fg_idx].mean(0)
    if len(bg_idx) > 0:
        click_f = click_f - bg_factor * affinity[bg_idx].mean(0)
    click_affinity = affinity_from_features(click_f.unsqueeze(1), distance=distance, affinity_focal_gamma=gamma)
    click_affinity = normalize_affinity(click_affinity)
    
    _A = bias_factor * click_affinity + (1 - bias_factor) * affinity
        
    eigvecs, eigvals, _ = svd_lowrank(_A, q=num_eig)
    eigvecs = correct_rotation(eigvecs)

    # propagate the eigenvectors to the full graph
    eigvecs = propagate_knn(eigvecs, features, features[fps_idx], distance=distance, device=device)
        
    return eigvecs, eigvals



def get_mask_and_heatmap(eigvecs, click_idx, num_cluster=2, device=None):
    device = which_device(eigvecs.device, device)
    eigvecs = eigvecs[:, :num_cluster].to(device)

    eigvecs = kway_ncut(eigvecs, return_continuous=True)
    # find which cluster is the foreground
    fg_eigvecs = eigvecs[click_idx]
    fg_idx = fg_eigvecs.mean(0).argmax().item()
    bg_idx = 1 if fg_idx == 0 else 0
    
    # discretize the eigvecs
    mask = eigvecs.argmax(dim=-1) == fg_idx

    heatmap = eigvecs[:, fg_idx] - eigvecs[:, bg_idx]
    
    return mask, heatmap


import logging
from typing import Any, Callable, Dict, Literal, Tuple

import numpy as np
import torch
import torch.nn.functional as F
from sklearn.base import BaseEstimator

from .nystrom_utils import (
    run_subgraph_sampling,
    propagate_knn,
    which_device,
)
from .math_utils import (
    quantile_normalize,
    quantile_min_max,
    check_if_normalized,
)


def _identity(X: torch.Tensor) -> torch.Tensor:
    return X


def _rgb_with_dimensionality_reduction(
    features: torch.Tensor,
    num_sample: int,
    metric: Literal["cosine", "euclidean"],
    rgb_func: Callable[[torch.Tensor, float], torch.Tensor],
    q: float, 
    knn: int,
    seed: int, 
    device: str,
    reduction: Callable[..., BaseEstimator],
    reduction_dim: int,
    reduction_kwargs: Dict[str, Any],
    transform_func: Callable[[torch.Tensor], torch.Tensor] = _identity,
) -> Tuple[torch.Tensor, torch.Tensor]:
    subgraph_indices = run_subgraph_sampling(
        features,
        num_sample=num_sample,
        sample_method="farthest",
    )

    _inp = features[subgraph_indices].cpu().numpy()
    _subgraph_embed = reduction(
        n_components=reduction_dim,
        metric=metric,
        random_state=seed,
        **reduction_kwargs
    ).fit_transform(_inp)

    _subgraph_embed = torch.tensor(_subgraph_embed, dtype=torch.float32)
    X_nd = transform_func(propagate_knn(
        _subgraph_embed,
        features,
        features[subgraph_indices],
        distance=metric,
        knn=knn,
        device=device,
        move_output_to_cpu=True,
    ))
    rgb = rgb_func(X_nd, q)
    return X_nd.numpy(force=True), rgb


def rgb_from_mspace_2d(
    features: torch.Tensor,
    q: float = 0.95,
    n_eig: int = 32,
    training_steps: int = 500,
    progress_bar: bool = True,
    **kwargs: Any,
):
    from .mspace import mspace_viz_transform
    """
    Returns:
        (torch.Tensor): Embedding in 2D, shape (n_samples, 2)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """

    x2d = mspace_viz_transform(features, n_eig=n_eig, mood_dim=2, training_steps=training_steps, progress_bar=progress_bar, **kwargs)

    rgb = rgb_from_2d_colormap(x2d, q=q)

    return x2d, rgb


def rgb_from_mspace_3d(
    features: torch.Tensor,
    q: float = 0.95,
    n_eig: int = 32,
    training_steps: int = 500,
    progress_bar: bool = True,
    **kwargs: Any,
):
    from .mspace import mspace_viz_transform
    """
    Returns:
        (torch.Tensor): Embedding in 3D, shape (n_samples, 3)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """

    x3d = mspace_viz_transform(features, n_eig=n_eig, mood_dim=3, training_steps=training_steps, progress_bar=progress_bar, **kwargs)

    rgb = rgb_from_3d_rgb_cube(x3d, q=q)

    return x3d, rgb



def rgb_from_tsne_2d(
    features: torch.Tensor,
    num_sample: int = 1000,
    perplexity: int = 150,
    metric: Literal["cosine", "euclidean"] = "cosine",
    device: str = None,
    seed: int = 0,
    q: float = 0.95,
    knn: int = 10,
    **kwargs: Any,
):
    """
    Returns:
        (torch.Tensor): Embedding in 2D, shape (n_samples, 2)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """
    try:
        from sklearn.manifold import TSNE
    except ImportError:
        raise ImportError(
            "sklearn import failed, please install `pip install scikit-learn`"
        )
    num_sample = min(num_sample, features.shape[0])
    if perplexity > num_sample // 2:
        logging.warning(
            f"perplexity is larger than num_sample, set perplexity to {num_sample // 2}"
        )
        perplexity = num_sample // 2

    x2d, rgb = _rgb_with_dimensionality_reduction(
        features=features,
        num_sample=num_sample,
        metric=metric,
        rgb_func=rgb_from_2d_colormap,
        q=q, knn=knn,
        seed=seed, device=device,
        reduction=TSNE, reduction_dim=2, reduction_kwargs={
            "perplexity": perplexity,
        },
    )
    
    return x2d, rgb


def rgb_from_tsne_3d(
    features: torch.Tensor,
    num_sample: int = 1000,
    perplexity: int = 150,
    metric: Literal["cosine", "euclidean"] = "cosine",
    device: str = None,
    seed: int = 0,
    q: float = 0.95,
    knn: int = 10,
    **kwargs: Any,
):
    """
    Returns:
        (torch.Tensor): Embedding in 3D, shape (n_samples, 3)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """
    try:
        from sklearn.manifold import TSNE
    except ImportError:
        raise ImportError(
            "sklearn import failed, please install `pip install scikit-learn`"
        )
    num_sample = min(num_sample, features.shape[0])
    if perplexity > num_sample // 2:
        logging.warning(
            f"perplexity is larger than num_sample, set perplexity to {num_sample // 2}"
        )
        perplexity = num_sample // 2

    x3d, rgb = _rgb_with_dimensionality_reduction(
        features=features,
        num_sample=num_sample,
        metric=metric,
        rgb_func=rgb_from_3d_rgb_cube,
        q=q, knn=knn,
        seed=seed, device=device,
        reduction=TSNE, reduction_dim=3, reduction_kwargs={
            "perplexity": perplexity,
        },
    )

    return x3d, rgb


def rgb_from_cosine_tsne_3d(
    features: torch.Tensor,
    num_sample: int = 1000,
    perplexity: int = 150,
    device: str = None,
    seed: int = 0,
    q: float = 0.95,
    knn: int = 10,
    **kwargs: Any,
):
    """
    Returns:
        (torch.Tensor): Embedding in 3D, shape (n_samples, 3)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """
    try:
        from sklearn.manifold import TSNE
    except ImportError:
        raise ImportError(
            "sklearn import failed, please install `pip install scikit-learn`"
        )
    num_sample = min(num_sample, features.shape[0])
    if perplexity > num_sample // 2:
        logging.warning(
            f"perplexity is larger than num_sample, set perplexity to {num_sample // 2}"
        )
        perplexity = num_sample // 2


    def cosine_to_rbf(X: torch.Tensor) -> torch.Tensor:                                 # [B... x N x 3]
        normalized_X = X / torch.norm(X, p=2, dim=-1, keepdim=True)                     # [B... x N x 3]
        D = 1 - normalized_X @ normalized_X.mT                                          # [B... x N x N]

        G = (D[..., :1, 1:] ** 2 + D[..., 1:, :1] ** 2 - D[..., 1:, 1:] ** 2) / 2       # [B... x (N - 1) x (N - 1)]
        L, V = torch.linalg.eigh(G)                                                     # [B... x (N - 1)], [B... x (N - 1) x (N - 1)]
        sqrtG = V[..., -3:] * (L[..., None, -3:] ** 0.5)                                # [B... x (N - 1) x 3]

        Y = torch.cat((torch.zeros_like(sqrtG[..., :1, :]), sqrtG), dim=-2)             # [B... x N x 3]
        Y = Y - torch.mean(Y, dim=-2, keepdim=True)
        return Y

    def rgb_from_cosine(X_3d: torch.Tensor, q: float) -> torch.Tensor:
        return rgb_from_3d_rgb_cube(cosine_to_rbf(X_3d), q=q)

    x3d, rgb = _rgb_with_dimensionality_reduction(
        features=features,
        num_sample=num_sample,
        metric="cosine",
        rgb_func=rgb_from_cosine,
        q=q, knn=knn,
        seed=seed, device=device,
        reduction=TSNE, reduction_dim=3, reduction_kwargs={
            "perplexity": perplexity,
        },
    )
    
    return x3d, rgb


def rgb_from_umap_2d(
    features: torch.Tensor,
    num_sample: int = 1000,
    n_neighbors: int = 150,
    min_dist: float = 0.1,
    metric: Literal["cosine", "euclidean"] = "cosine",
    device: str = None,
    seed: int = 0,
    q: float = 0.95,
    knn: int = 10,
    **kwargs: Any,
):
    """
    Returns:
        (torch.Tensor): Embedding in 2D, shape (n_samples, 2)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """
    try:
        from umap import UMAP
    except ImportError:
        raise ImportError("umap import failed, please install `pip install umap-learn`")

    x2d, rgb = _rgb_with_dimensionality_reduction(
        features=features,
        num_sample=num_sample,
        metric=metric,
        rgb_func=rgb_from_2d_colormap,
        q=q, knn=knn,
        seed=seed, device=device,
        reduction=UMAP, reduction_dim=2, reduction_kwargs={
            "n_neighbors": n_neighbors,
            "min_dist": min_dist,
        },
    )
    
    return x2d, rgb


def rgb_from_umap_sphere(
    features: torch.Tensor,
    num_sample: int = 1000,
    n_neighbors: int = 150,
    min_dist: float = 0.1,
    metric: Literal["cosine", "euclidean"] = "cosine",
    device: str = None,
    seed: int = 0,
    q: float = 0.95,
    knn: int = 10,
    **kwargs: Any,
):
    """
    Returns:
        (torch.Tensor): Embedding in 2D, shape (n_samples, 2)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """
    try:
        from umap import UMAP
    except ImportError:
        raise ImportError("umap import failed, please install `pip install umap-learn`")

    def transform_func(X: torch.Tensor) -> torch.Tensor:
        return torch.stack((
            torch.sin(X[:, 0]) * torch.cos(X[:, 1]),
            torch.sin(X[:, 0]) * torch.sin(X[:, 1]),
            torch.cos(X[:, 0]),
        ), dim=1)

    x3d, rgb = _rgb_with_dimensionality_reduction(
        features=features,
        num_sample=num_sample,
        metric=metric,
        rgb_func=rgb_from_3d_rgb_cube,
        q=q, knn=knn,
        seed=seed, device=device,
        reduction=UMAP, reduction_dim=2, reduction_kwargs={
            "n_neighbors": n_neighbors,
            "min_dist": min_dist,
            "output_metric": "haversine",
        },
        transform_func=transform_func
    )
    
    return x3d, rgb


def rgb_from_umap_3d(
    features: torch.Tensor,
    num_sample: int = 1000,
    n_neighbors: int = 150,
    min_dist: float = 0.1,
    metric: Literal["cosine", "euclidean"] = "cosine",
    device: str = None,
    seed: int = 0,
    q: float = 0.95,
    knn: int = 10,
    **kwargs: Any,
):
    """
    Returns:
        (torch.Tensor): Embedding in 2D, shape (n_samples, 2)
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """
    try:
        from umap import UMAP
    except ImportError:
        raise ImportError("umap import failed, please install `pip install umap-learn`")

    x3d, rgb = _rgb_with_dimensionality_reduction(
        features=features,
        num_sample=num_sample,
        metric=metric,
        rgb_func=rgb_from_3d_rgb_cube,
        q=q, knn=knn,
        seed=seed, device=device,
        reduction=UMAP, reduction_dim=3, reduction_kwargs={
            "n_neighbors": n_neighbors,
            "min_dist": min_dist,
        },
    )
    
    return x3d, rgb


def flatten_sphere(X_3d):
    x = np.arctan2(X_3d[:, 0], X_3d[:, 1])
    y = -np.arccos(X_3d[:, 2])
    X_2d = np.stack([x, y], axis=1)
    return X_2d


def rotate_rgb_cube(rgb, position=1):
    """rotate RGB cube to different position

    Args:
        rgb (torch.Tensor): RGB color space [0, 1], shape (*, 3)
        position (int): position to rotate, 0, 1, 2, 3, 4, 5, 6

    Returns:
        torch.Tensor: RGB color space, shape (n_samples, 3)
    """
    assert position in range(0, 7), "position should be 0, 1, 2, 3, 4, 5, 6"
    rotation_matrix = torch.tensor(
        [
            [0, 1, 0],
            [0, 0, 1],
            [1, 0, 0],
        ]
    ).float()
    n_mul = position % 3
    rotation_matrix = torch.matrix_power(rotation_matrix, n_mul)
    rgb = rgb @ rotation_matrix
    if position > 3:
        rgb = 1 - rgb
    return rgb


def rgb_from_3d_rgb_cube(X_3d, q=0.95):
    """convert 3D t-SNE to RGB color space

    Args:
        X_3d (torch.Tensor): 3D t-SNE embedding, shape (n_samples, 3)
        q (float): quantile, default 0.95

    Returns:
        torch.Tensor: RGB color space, shape (n_samples, 3)
    """
    assert X_3d.shape[1] == 3, "input should be (n_samples, 3)"
    assert len(X_3d.shape) == 2, "input should be (n_samples, 3)"
    rgb = []
    for i in range(3):
        rgb.append(quantile_normalize(X_3d[:, i], q=q))
    rgb = torch.stack(rgb, dim=-1)
    return rgb


def convert_to_lab_color(rgb, full_range=True):
    from skimage import color
    import copy

    if isinstance(rgb, torch.Tensor):
        rgb = rgb.cpu().numpy()
    _rgb = copy.deepcopy(rgb)
    _rgb[..., 0] = _rgb[..., 0] * 100
    if full_range:
        _rgb[..., 1] = _rgb[..., 1] * 255 - 128
        _rgb[..., 2] = _rgb[..., 2] * 255 - 128
    else:
        _rgb[..., 1] = _rgb[..., 1] * 100 - 50
        _rgb[..., 2] = _rgb[..., 2] * 100 - 50
    lab_rgb = color.lab2rgb(_rgb)
    return lab_rgb


def rgb_from_2d_colormap(X_2d, q=0.95):
    xy = X_2d.clone()
    for i in range(2):
        xy[:, i] = quantile_normalize(xy[:, i], q=q)

    try:
        from pycolormap_2d import (
            ColorMap2DBremm,
            ColorMap2DZiegler,
            ColorMap2DCubeDiagonal,
            ColorMap2DSchumann,
        )
    except ImportError:
        raise ImportError(
            "pycolormap_2d import failed, please install `pip install pycolormap-2d`"
        )

    cmap = ColorMap2DCubeDiagonal()
    xy = xy.cpu().numpy()
    len_x, len_y = cmap._cmap_data.shape[:2]
    x = (xy[:, 0] * (len_x - 1)).astype(np.int64)
    y = (xy[:, 1] * (len_y - 1)).astype(np.int64)
    x = np.clip(x, 0, len_x - 1)
    y = np.clip(y, 0, len_y - 1)
    rgb = cmap._cmap_data[x, y]
    rgb = torch.tensor(rgb, dtype=torch.float32) / 255
    return rgb

def rgb_from_nd_colormap(X_nd, q=0.95, lab_color=False):
    """
    Returns:
        (torch.Tensor): RGB color for each data sample, shape (n_samples, 3)
    """
    d = X_nd.shape[1]
    if d == 2:
        return rgb_from_2d_colormap(X_nd, q=q)
    elif d == 3:
        rgb = rgb_from_3d_rgb_cube(X_nd, q=q)
        if lab_color:
            rgb = convert_to_lab_color(rgb)
            rgb = torch.from_numpy(rgb)
        return rgb
    else:
        raise ValueError(f"Unsupported dimensionality: {d}")



# %%
import logging
import math
from typing import Literal

import numpy as np
import torch
import torch.nn.functional as F

from .nystrom_utils import (
    run_subgraph_sampling,
    propagate_knn,
    farthest_point_sampling,
    which_device,
)
from .math_utils import (
    affinity_from_features,
    normalize_affinity,
    svd_lowrank,
    gram_schmidt,
    correct_rotation,
)
from .affinity_gamma import find_gamma_by_degree_after_fps


def nystrom_ncut(
    features: torch.Tensor,
    num_eig: int = 100,
    degree: float = 0.05,
    distance: Literal["cosine", "euclidean", "rbf"] = "rbf",
    num_sample: int = 10240,
    num_sample2: int = 1024,
    knn: int = 10,
    matmul_chunk_size: int = 16384,
    sample_method: Literal["farthest", "random"] = "farthest",
    precomputed_sampled_indices: torch.Tensor = None,
    affinity_focal_gamma: float = None,
    device: str = None,
    make_orthogonal: bool = False,
    no_propagation: bool = False,
    move_output_to_cpu: bool = None,
    **kwargs,
):
    """PyTorch implementation of Faster Nystrom Normalized cut.
    Args:
        features (torch.Tensor): feature matrix, shape (n_samples, n_features)
        num_eig (int): default 100, number of top eigenvectors to return
        degree (float): target degree to search for optimal gamma, default 0.05. 
            lower degree will result in more sharp eigenvectors
        distance (str): distance metric, 'cosine' (default) or 'euclidean', 'rbf'
        num_sample (int): default 10240, number of samples for Nystrom-like approximation
        num_sample2 (int): default 1024, number of samples for eigenvector propagation
        knn (int): default 10, number of KNN for propagating eigenvectors from subgraph to full graph,
            smaller knn will result in more sharp eigenvectors,
        sample_method (str): sample method, 'farthest' (default) or 'random'
            'farthest' is recommended for better approximation
        precomputed_sampled_indices (torch.Tensor): precomputed sampled indices, shape (num_sample,)
            override the sample_method, if not None
        affinity_focal_gamma (float): affinity matrix parameter, lower t reduce the weak edge weights,
            resulting in more sharp eigenvectors, default None (auto search)
        device (str): device to use for computation, if None, will not change device
            a good practice is to pass features by CPU since it's usually large,
            and move subgraph affinity to GPU to speed up eigenvector computation
        make_orthogonal (bool): make eigenvectors orthogonal after propagation, default True
        no_propagation (bool): if True, skip the eigenvector propagation step, only return the subgraph eigenvectors
        move_output_to_cpu (bool): move output to CPU, set to True if output is too large to fit in GPU memory
    Returns:
        (torch.Tensor): eigenvectors, shape (n_samples, num_eig)
        (torch.Tensor): eigenvalues, sorted in descending order, shape (num_eig,)
        (torch.Tensor): sampled_indices used by Nystrom-like approximation subgraph, shape (num_sample,)
    Examples:
        >>> from ncut_pytorch import nystrom_ncut
        >>> import torch
        >>> features = torch.rand(10000, 100)
        >>> eigenvectors, eigenvalues = nystrom_ncut(features, num_eig=20)
        >>> print(eigenvectors.shape, eigenvalues.shape)
        >>> # (10000, 20) (20,)
    """

    if precomputed_sampled_indices is not None:
        sampled_indices = precomputed_sampled_indices
    else:
        sampled_indices = run_subgraph_sampling(
            features,
            num_sample=num_sample,
            sample_method=sample_method,
        )

    if move_output_to_cpu is None:
        move_output_to_cpu = True if features.device.type == "cpu" else False

    sampled_features = features[sampled_indices]
    device = which_device(sampled_features.device, device)
    sampled_features = sampled_features.to(device)

    # compute affinity matrix on subgraph
    if affinity_focal_gamma is None:
        with torch.no_grad():
            affinity_focal_gamma = find_gamma_by_degree_after_fps(sampled_features, degree, distance=distance)
    
    A = affinity_from_features(sampled_features, affinity_focal_gamma=affinity_focal_gamma, distance=distance)

    # compute normalized cut on the subgraph
    eigen_vector, eigen_value = ncut(A, num_eig)

    if no_propagation:
        return eigen_vector, eigen_value, sampled_indices, affinity_focal_gamma

    # propagate eigenvectors from subgraph to full graph
    eigen_vector = propagate_knn(
        eigen_vector,
        features,
        features[sampled_indices],
        knn,
        num_sample=num_sample2,
        distance=distance,
        affinity_focal_gamma=affinity_focal_gamma,
        chunk_size=matmul_chunk_size,
        device=device,
        move_output_to_cpu=move_output_to_cpu,
    )

    # post-hoc orthogonalization
    if make_orthogonal:
        eigen_vector = gram_schmidt(eigen_vector)

    return eigen_vector, eigen_value


class NCUT:
    """Nystrom Normalized Cut."""

    def __init__(
        self,
        num_eig: int = 100,
        knn: int = 10,
        affinity_focal_gamma: float = None,
        degree: float = 0.05,
        num_sample: int = 10240,
        num_sample2: int = 1024,
        sample_method: Literal["farthest", "random"] = "farthest",
        distance: Literal["cosine", "euclidean", "rbf"] = "rbf",
        device: str = None,
        move_output_to_cpu: bool = False,
        make_orthogonal: bool = False,
        **kwargs,
    ):
        """
        Args:
            num_eig (int): default 100, number of top eigenvectors to return
            num_sample (int): default 10240, number of samples for Nystrom-like approximation
            num_sample2 (int): default 1024, number of samples for eigenvector propagation
            knn (int): default 10, number of KNN for propagating eigenvectors from subgraph to full graph,
                smaller knn will result in more sharp eigenvectors,
            sample_method (str): sample method, 'farthest' (default) or 'random'
                'farthest' is recommended for better approximation
            distance (str): distance metric, 'cosine' (default) or 'euclidean', 'rbf'
            affinity_focal_gamma (float): affinity matrix parameter, lower t reduce the weak edge weights,
                resulting in more sharp eigenvectors, default None (auto search)
            degree (float): target degree to search for optimal gamma, default 0.05. 
                lower degree will result in more sharp eigenvectors
            device (str): device to use for computation, if None, will not change device
                a good practice is to pass features by CPU since it's usually large,
                and move subgraph affinity to GPU to speed up eigenvector computation
            make_orthogonal (bool): make eigenvectors orthogonal after propagation, default True
            move_output_to_cpu (bool): move output to CPU, set to True if you have memory issue

        Examples:
            >>> from ncut_pytorch import NCUT
            >>> import torch
            >>> features = torch.rand(10000, 100)
            >>> ncut = NCUT(num_eig=20)
            >>> ncut.fit(features)
            >>> eigenvectors, eigenvalues = ncut.transform(features)
            >>> print(eigenvectors.shape, eigenvalues.shape)
            >>> # (10000, 20) (20,)

            >>> # transform new features
            >>> new_features = torch.rand(500, 100)
            >>> new_eigenvectors, _ = ncut.transform(new_features)
            >>> print(new_eigenvectors.shape)
            >>> # (500, 20)
        """
        self.num_eig = num_eig
        self.num_sample = num_sample
        self.num_sample2 = num_sample2
        self.knn = knn
        self.sample_method = sample_method
        self.distance = distance
        self.affinity_focal_gamma = affinity_focal_gamma
        self.degree = degree
        self.device = device
        self.move_output_to_cpu = move_output_to_cpu
        self.make_orthogonal = make_orthogonal

        self.subgraph_eigen_vector = None
        self.eigen_value = None
        self.subgraph_sample_indices = None
        self.subgraph_features = None

    def fit(self,
            features: torch.Tensor,
            precomputed_sampled_indices: torch.Tensor = None
            ):
        """Fit Nystrom Normalized Cut on the input features.
        Args:
            features (torch.Tensor): input features, shape (n_samples, n_features)
            precomputed_sampled_indices (torch.Tensor): precomputed sampled indices, shape (num_sample,)
                override the sample_method, if not None
        Returns:
            (NCUT): self
        """
        _n = features.shape[0]
        if self.num_sample >= _n:
            logging.info(
                f"NCUT nystrom num_sample is larger than number of input samples, nyström approximation is not needed, setting num_sample={_n} and knn=1"
            )
            self.num_sample = _n
            self.knn = 1

        # save the eigenvectors solution on the sub-sampled graph, do not propagate to full graph yet
        (self.subgraph_eigen_vector, self.eigen_value,
        self.subgraph_sample_indices, self.affinity_focal_gamma) = nystrom_ncut(
                                features,
                                num_eig=self.num_eig,
                                num_sample=self.num_sample,
                                num_sample2=self.num_sample2,
                                sample_method=self.sample_method,
                                precomputed_sampled_indices=precomputed_sampled_indices,
                                distance=self.distance,
                                affinity_focal_gamma=self.affinity_focal_gamma,
                                degree=self.degree,
                                device=self.device,
                                move_output_to_cpu=self.move_output_to_cpu,
                                no_propagation=True,
                            )
        self.subgraph_features = features[self.subgraph_sample_indices]
        return self

    def transform(self, features: torch.Tensor):
        """Transform new features using the fitted Nystrom Normalized Cut.
        Args:
            features (torch.Tensor): new features, shape (n_samples, n_features)
        Returns:
            (torch.Tensor): eigen_vectors, shape (n_samples, num_eig)
            (torch.Tensor): eigen_values, sorted in descending order, shape (num_eig,)
        """
    
        # propagate eigenvectors from subgraph to full graph
        eigen_vector = propagate_knn(
            self.subgraph_eigen_vector,
            features,
            self.subgraph_features,
            self.knn,
            num_sample=self.num_sample2,
            distance=self.distance,
            affinity_focal_gamma=self.affinity_focal_gamma,
            device=self.device,
            move_output_to_cpu=self.move_output_to_cpu,
        )
        if self.make_orthogonal:
            eigen_vector = gram_schmidt(eigen_vector)
        return eigen_vector, self.eigen_value

    def fit_transform(self,
                      features: torch.Tensor,
                      precomputed_sampled_indices: torch.Tensor = None
                      ):
        """
        Args:
            features (torch.Tensor): input features, shape (n_samples, n_features)
            precomputed_sampled_indices (torch.Tensor): precomputed sampled indices, shape (num_sample,)
                override the sample_method, if not None
                
        Returns:
            (torch.Tensor): eigen_vectors, shape (n_samples, num_eig)
            (torch.Tensor): eigen_values, sorted in descending order, shape (num_eig,)
        """
        return self.fit(features, precomputed_sampled_indices=precomputed_sampled_indices).transform(features)


def ncut(
    A: torch.Tensor,
    num_eig: int = 100,
):
    """PyTorch implementation of Normalized cut without Nystrom-like approximation.

    Args:
        A (torch.Tensor): affinity matrix, shape (n_samples, n_samples)
        num_eig (int): number of eigenvectors to return

    Returns:
        (torch.Tensor): eigenvectors corresponding to the eigenvalues, shape (n_samples, num_eig)
        (torch.Tensor): eigenvalues of the eigenvectors, sorted in descending order
    """

    # normalization; A = D^(-1/2) A D^(-1/2)
    A = normalize_affinity(A)

    eigen_vector, eigen_value, _ = svd_lowrank(A, num_eig)

    # correct the random rotation (flipping sign) of eigenvectors
    eigen_vector = correct_rotation(eigen_vector)

    return eigen_vector, eigen_value

    

## dirty hack for backward compatibility ##

try:
    from .math_utils import (
        quantile_normalize,
        quantile_min_max,
    )
    from .visualize_utils import (
        rgb_from_tsne_3d,
        rgb_from_umap_sphere,
        rgb_from_tsne_2d,
        rgb_from_umap_3d,
        rgb_from_umap_2d,
        rotate_rgb_cube,
        convert_to_lab_color,
    )
    from .kway_ncut import kway_ncut, axis_align
except ImportError:
    print("some of viualization and nystrom_utils are not imported")


import torch
import math
import numpy as np
from typing import Literal
from torch.nn import functional as F


def affinity_from_features(
    features: torch.Tensor,
    features_B: torch.Tensor = None,
    affinity_focal_gamma: float = 1.0,
    distance: Literal["cosine", "euclidean", "rbf"] = "rbf",
    **kwargs,
):
    """Compute affinity matrix from input features.

    Args:
        features (torch.Tensor): input features, shape (n_samples, n_features)
        feature_B (torch.Tensor, optional): optional, if not None, compute affinity between two features
        affinity_focal_gamma (float): affinity matrix parameter, lower t reduce the edge weights
            on weak connections, default 1.0
        distance (str): distance metric, 'cosine' (default) or 'euclidean', 'rbf'.
        normalize_features (bool): normalize input features before computing affinity matrix

    Returns:
        (torch.Tensor): affinity matrix, shape (n_samples, n_samples)
    """
    # compute affinity matrix from input features

    features_B = features if features_B is None else features_B

    if distance == "cosine":
        features = F.normalize(features, dim=-1)
        features_B = F.normalize(features_B, dim=-1)
        A = 1 - features @ features_B.T
    elif distance == "euclidean":
        A = torch.cdist(features, features_B, p=2)
    elif distance == "rbf":
        d = torch.cdist(features, features_B, p=2)
        A = torch.pow(d, 2)
    else:
        raise ValueError("distance should be 'cosine' or 'euclidean', 'rbf'")

    if distance == "rbf":
        sigma = 2 * affinity_focal_gamma * features.var(dim=0).sum()
    else:
        sigma = affinity_focal_gamma

    A = torch.exp(-A / sigma)

    return A

def svd_lowrank(mat, q):
    """
    SVD lowrank
    mat: (n, m), n data, m features
    q: int
    return: (n, q), (q,), (q, m)
    """
    u, s, v = torch.svd_lowrank(mat, q=q+10)
    # take 10 extra components to reduce the error, because error in svd_lowrank
    u = u[:, :q]
    s = s[:q]
    v = v[:, :q]
    return u, s, v


def pca_lowrank(mat, q):
    """
    PCA lowrank
    mat: (n, m), n data, m features
    q: int
    return: (n, q), (q,), (q, m)
    """
    u, s, v = svd_lowrank(mat, q)
    _n = mat.shape[0]
    s /= math.sqrt(_n)
    return u @ torch.diag(s) 


def check_if_normalized(x, n_sample=1000):
    """check if the input tensor is normalized (unit norm)"""
    n_sample = min(n_sample, x.shape[0])
    random_indices = torch.randperm(x.shape[0])[:n_sample]
    _x = x[random_indices]
    flag = torch.allclose(torch.norm(_x, dim=-1), torch.ones(n_sample, device=x.device))
    return flag


def quantile_min_max(x, q1=0.01, q2=0.99, n_sample=10000):
    if x.shape[0] > n_sample:
        # random sampling to reduce the load of quantile calculation, torch.quantile does not support large tensor
        np.random.seed(0)
        random_idx = np.random.choice(x.shape[0], n_sample, replace=False)
        vmin, vmax = x[random_idx].quantile(q1), x[random_idx].quantile(q2)
    else:
        vmin, vmax = x.quantile(q1), x.quantile(q2)
    return vmin, vmax


def quantile_normalize(x, q=0.95):
    """normalize each dimension of x to [0, 1], take 95-th percentage, this robust to outliers
        </br> 1. sort x
        </br> 2. take q-th quantile
        </br>     min_value -> (1-q)-th quantile
        </br>     max_value -> q-th quantile
        </br> 3. normalize
        </br> x = (x - min_value) / (max_value - min_value)

    Args:
        x (torch.Tensor): input tensor, shape (n_samples, n_features)
            normalize each feature to 0-1 range
        q (float): quantile, default 0.95

    Returns:
        torch.Tensor: quantile normalized tensor
    """
    # normalize x to 0-1 range, max value is q-th quantile
    # quantile makes the normalization robust to outliers
    if isinstance(x, np.ndarray):
        x = torch.tensor(x)
    vmax, vmin = quantile_min_max(x, q, 1 - q)
    x = (x - vmin) / (vmax - vmin)
    x = x.clamp(0, 1)
    return x


def gram_schmidt(matrix):
    """Orthogonalize a matrix column-wise using the Gram-Schmidt process.

    Args:
        matrix (torch.Tensor): A matrix to be orthogonalized (m x n).
            the second dimension is orthogonalized
    Returns:
        torch.Tensor: Orthogonalized matrix (m x n).
    """

    # Get the number of rows (m) and columns (n) of the input matrix
    m, n = matrix.shape

    # Create an empty matrix to store the orthogonalized columns
    orthogonal_matrix = torch.zeros((m, n), dtype=matrix.dtype, device=matrix.device)

    for i in range(n):
        # Start with the i-th column of the input matrix
        vec = matrix[:, i]

        for j in range(i):
            # Subtract the projection of vec onto the j-th orthogonal column
            proj = torch.dot(orthogonal_matrix[:, j], matrix[:, i]) / torch.dot(
                orthogonal_matrix[:, j], orthogonal_matrix[:, j]
            )
            vec = vec - proj * orthogonal_matrix[:, j]

        # Store the orthogonalized vector
        orthogonal_matrix[:, i] = vec / torch.norm(vec)

    return orthogonal_matrix


def chunked_matmul(
    A,
    B,
    chunk_size=8096,
    device="cuda:0",
    large_device="cpu",
    transform=lambda x: x,
):
    A = A.to(large_device)
    B = B.to(large_device)
    C = torch.zeros(A.shape[0], B.shape[1], device=large_device)
    iterator = range(0, A.shape[0], chunk_size)
    for i in iterator:
        end_i = min(i + chunk_size, A.shape[0])
        for j in range(0, B.shape[1], chunk_size):
            end_j = min(j + chunk_size, B.shape[1])
            _A = A[i:end_i]
            _B = B[:, j:end_j]
            _C_ij = None
            for k in range(0, A.shape[1], chunk_size):
                end_k = min(k + chunk_size, A.shape[1])
                __A = _A[:, k:end_k].to(device)
                __B = _B[k:end_k].to(device)
                _C = __A @ __B
                _C_ij = _C if _C_ij is None else _C_ij + _C
            _C_ij = transform(_C_ij)

            _C_ij = _C_ij.to(large_device)
            C[i:end_i, j:end_j] = _C_ij
    return C


def correct_rotation(eigen_vector):
    # correct the random rotation (flipping sign) of eigenvectors
    with torch.no_grad():
        rand_w = torch.ones(
            eigen_vector.shape[0], device=eigen_vector.device, dtype=eigen_vector.dtype
        )
        s = rand_w[None, :] @ eigen_vector
        s = s.sign()
    eigen_vector = eigen_vector * s
    return eigen_vector


def normalize_affinity(A, eps=1e-8):
    with torch.no_grad():
        D = A.abs().sum(dim=-1)
        D = D + eps
    A = A / torch.sqrt(D)[:, None]
    A = A / torch.sqrt(D)[None, :]
    return A


@torch.no_grad()
def compute_delaunay(points):
    """Compute Delaunay triangulation of points"""
    from scipy.spatial import Delaunay
    if points.shape[1] > 3:
        points = pca_lowrank(points, 3)   # TODO: does reduce to 2d make it a plane even if mspace is 3d?
    return Delaunay(points.cpu().numpy()).simplices
  
  
def compute_riemann_curvature_loss(points, simplices=None, domain_min=0, domain_max=1):
    """
    Calculate loss based on approximated Riemann curvature.
    
    The loss measures deviations from uniform metric tensors across simplices,
    which approximates variations in Riemann curvature.
    """
    if simplices is None:
        simplices = compute_delaunay(points)
        
    ideal_det = torch.tensor(1.0, device=points.device, dtype=torch.float64)
    
    # Process each simplex in parallel 
    simplices_tensor = torch.tensor(simplices, device=points.device)
    
    # Extract points that form each simplex
    simplex_points = points[simplices_tensor]
    
    # Calculate edge vectors from the first point of each simplex
    edges = simplex_points[:, 1:] - simplex_points[:, 0].unsqueeze(1)
    
    # Compute metric tensors (Gram matrices) for each simplex
    metric_tensors = torch.matmul(edges, edges.transpose(1, 2))
    
    # Calculate determinants (related to volume distortion)
    dets = torch.linalg.det(metric_tensors)
    
    # Penalize deviations from constant determinant
    valid_dets = dets[dets > 0]
    total_curvature = torch.mean((valid_dets - ideal_det)**2)
    return total_curvature


def compute_axis_align_loss(data):
    """ Encourage axis alignment by minimizing off-diagonal elements in the covariance matrix """
    n, d = data.shape
    centered_data = data - data.mean(dim=0)  # Center the data
    cov_matrix = (centered_data.T @ centered_data) / n  # Compute covariance matrix
    
    eye = torch.eye(d, device=data.device)
    return torch.nn.functional.smooth_l1_loss(cov_matrix, eye)  # L1 loss between covariance matrix and identity matrix


def compute_repulsion_loss(points):

    # Add point repulsion term for additional stability
    dist_matrix = torch.cdist(points, points)
    # Set diagonal to large value to avoid self-repulsion
    mask = torch.eye(points.shape[0], device=points.device).bool()
    dist_matrix = dist_matrix + mask * 1000
    repulsion = 1.0 / (dist_matrix + 1e-8)
    non_diag = repulsion[~mask]
    return torch.mean(non_diag)

def compute_boundary_loss(points, domain_min=-1, domain_max=1):
    return torch.mean(torch.relu(domain_min - points)) + \
            torch.mean(torch.relu(points - domain_max))


from collections import defaultdict
import logging
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import pytorch_lightning as pl
from tqdm import tqdm
from torch.utils.data import TensorDataset

# disable lightning logs
logging.getLogger('lightning').setLevel(0)
logging.getLogger("pytorch_lightning").setLevel(0)
class IgnorePLFilter(logging.Filter):
    def filter(self, record):
        keywords = ['available:', 'CUDA', 'LOCAL_RANK:']
        return not any(keyword in record.getMessage() for keyword in keywords)
logging.getLogger('pytorch_lightning.utilities.rank_zero').addFilter(IgnorePLFilter())
logging.getLogger('pytorch_lightning.accelerators.cuda').addFilter(IgnorePLFilter())


from .ncut_pytorch import affinity_from_features, ncut
from .affinity_gamma import find_gamma_by_degree_after_fps
from .math_utils import compute_riemann_curvature_loss, compute_boundary_loss, compute_repulsion_loss, compute_axis_align_loss

def _kway_ncut_loss(eigvec_gt, eigvec_hat, n_eig):
    _eigvec_gt = eigvec_gt[:, :n_eig]
    _eigvec_hat = eigvec_hat[:, :n_eig]
    loss = F.smooth_l1_loss(_eigvec_gt @ _eigvec_gt.T, _eigvec_hat @ _eigvec_hat.T)
    return loss

def flag_space_loss(eigvec_gt, eigvec_hat, n_eig, start=4, step_mult=2):
    if torch.all(eigvec_gt == 0) or torch.all(eigvec_hat == 0):
        return torch.tensor(0, device=eigvec_gt.device)
    
    loss = 0
    n_eig = start // step_mult
    while True:
        n_eig *= step_mult
        loss += _kway_ncut_loss(eigvec_gt, eigvec_hat, n_eig)
        if n_eig > eigvec_gt.shape[1] or n_eig > eigvec_hat.shape[1]:
            break
    return loss

def ncut_wrapper(features, n_eig, distance='rbf', gamma=0.5):
    A = affinity_from_features(features, distance=distance, gamma=gamma)
    eigvec, eigval = ncut(A, n_eig)
    return eigvec, eigval


class MLP(nn.Module):
    def __init__(self, in_dim, out_dim, n_layer=2, latent_dim=4096):
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Linear(in_dim, latent_dim),
            nn.GELU(),
            *[nn.Sequential(nn.Linear(latent_dim, latent_dim), nn.GELU()) for _ in range(n_layer)],
            nn.Linear(latent_dim, out_dim)
        )
    
    def forward(self, x):
        return self.mlp(x)


class CompressionModel(pl.LightningModule):
    def __init__(self, in_dim, mood_dim=2, n_eig=32, 
                 n_layer=4, latent_dim=256, 
                 eigvec_loss=1, recon_loss=1, 
                 riemann_curvature_loss=0.1, axis_align_loss=0, 
                 repulsion_loss=0.1, boundary_loss=1, zero_center_loss=0.1,
                 lr=0.005, progress_bar=True, training_steps=500):
        super().__init__()
        
        self.compress = MLP(in_dim, mood_dim, n_layer, latent_dim)
        self.uncompress = MLP(mood_dim, in_dim, n_layer, latent_dim)
                
        self.loss_history = defaultdict(list)

        self.n_eig = n_eig
        self.eigvec_loss = eigvec_loss
        self.recon_loss = recon_loss
        self.riemann_curvature_loss = riemann_curvature_loss
        self.axis_align_loss = axis_align_loss
        self.repulsion_loss = repulsion_loss
        self.boundary_loss = boundary_loss
        self.zero_center_loss = zero_center_loss
        self.lr = lr
        
        self.progress_bar = progress_bar
        self.training_steps = training_steps
        if self.progress_bar:
            self.progress_bar = tqdm(total=training_steps, desc="Mspace training")


    def forward(self, x):
        if isinstance(x, list):
            x = x[0]
        return self.compress(x)

    def training_step(self, batch):
        if self.progress_bar and self.trainer.global_step % 10 == 0:
            self.progress_bar.update(10)
        if self.progress_bar and self.trainer.global_step >= self.training_steps - 1:
            self.progress_bar.close()

        input_feats, output_feats = batch

        if self.trainer.global_step == 0:
            self.gamma = find_gamma_by_degree_after_fps(input_feats, 0.1, distance='rbf')

        feats_compressed = self.compress(input_feats)
        feats_uncompressed = self.uncompress(feats_compressed)
        
        eigvec_gt, eigval_gt = ncut_wrapper(input_feats, self.n_eig, gamma=self.gamma, distance='rbf')
        eigvec_hat, eigval_hat = ncut_wrapper(feats_compressed, self.n_eig, gamma=self.gamma, distance='rbf')

        total_loss = 0
        if self.eigvec_loss > 0:
            eigvec_loss = flag_space_loss(eigvec_gt, eigvec_hat, n_eig=self.n_eig)
            self.log("loss/eigvec", eigvec_loss, prog_bar=True)
            total_loss += eigvec_loss * self.eigvec_loss
            self.loss_history['eigvec'].append(eigvec_loss.item())

        if self.recon_loss > 0:
            recon_loss = F.smooth_l1_loss(output_feats, feats_uncompressed)
            self.log("loss/recon", recon_loss, prog_bar=True)
            total_loss += recon_loss * self.recon_loss
            self.loss_history['recon'].append(recon_loss.item())

        if self.riemann_curvature_loss > 0:
            riemann_curvature_loss = compute_riemann_curvature_loss(feats_compressed)
            self.log("loss/riemann_curvature", riemann_curvature_loss, prog_bar=True)
            total_loss += riemann_curvature_loss * self.riemann_curvature_loss

        if self.axis_align_loss > 0:
            axis_align_loss = compute_axis_align_loss(feats_compressed)
            self.log("loss/axis_align", axis_align_loss, prog_bar=True)
            total_loss += axis_align_loss * self.axis_align_loss

        if self.repulsion_loss > 0:
            repulsion_loss = compute_repulsion_loss(feats_compressed)
            self.log("loss/repulsion", repulsion_loss, prog_bar=True)
            total_loss += repulsion_loss * self.repulsion_loss

        if self.boundary_loss > 0:
            boundary_loss = compute_boundary_loss(feats_compressed)
            self.log("loss/boundary", boundary_loss, prog_bar=True)
            total_loss += boundary_loss * self.boundary_loss

        if self.zero_center_loss > 0:
            zero_center_loss = feats_compressed.abs().mean()
            self.log("loss/zero_center", zero_center_loss, prog_bar=True)
            total_loss += zero_center_loss * self.zero_center_loss

        loss = total_loss
        self.log("loss/total", loss, prog_bar=True)
        return loss
    
    def configure_optimizers(self):
        optimizer = torch.optim.NAdam(self.parameters(), lr=self.lr)
        return optimizer


def train_mspace_model(compress_feats, uncompress_feats, training_steps=500, grad_clip_val=1.0, 
                    batch_size=1000, devices=[0], return_trainer=False, progress_bar=True, **model_kwargs):
    compress_feats = torch.tensor(compress_feats).float().cpu()
    uncompress_feats = torch.tensor(uncompress_feats).float().cpu()
    l, c = compress_feats.shape

    model = CompressionModel(c, training_steps=training_steps, progress_bar=progress_bar, **model_kwargs)
    
    dataset = TensorDataset(compress_feats, uncompress_feats)
    dataloader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=True)

    is_cuda = torch.cuda.is_available()
    trainer = pl.Trainer(limit_train_batches=training_steps,
                         max_steps=training_steps,
                         gradient_clip_val=grad_clip_val,
                         accelerator="gpu" if is_cuda else "cpu", 
                         devices=devices if is_cuda else None,
                         enable_checkpointing=False,
                         enable_progress_bar=False,
                         enable_model_summary=False,
                         logger=False,
                         )
    trainer.fit(model, dataloader)

    if return_trainer:
        return model, trainer

    return model

def mspace_viz_transform(feats, **kwargs):
    model, trainer = train_mspace_model(feats, feats, return_trainer=True, **kwargs)

    batch_size = kwargs.get('batch_size', 1000)
    test_loader = torch.utils.data.DataLoader(TensorDataset(feats), batch_size=batch_size, shuffle=False)
    compressed = trainer.predict(model, test_loader)
    compressed = torch.cat(compressed, dim=0)
    return compressed
